/*
   FILE:							       	
	rtl_tasks.c							       
								       	
   DESCRIPTION:							       	
	RT-Linux side of a sample application using the POSIX trace
	support.

	See README file for details.
					       				
   Copyright (C) 2001
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Ignacio Paches Giner
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        September 2001

*/

#include <rtl.h>
#include <time.h>
#include <pthread.h>
#include <rtl_sched.h>
#include <trace.h>

#include <metrics.h>

#define _DEBUG


static trace_id_t       trid;
static pthread_t        task1, task2, task3, task_reader;

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
static char            str[TRACE_NAME_MAX];


void *just_execute(void *task_period) {
  int       j;
  long      period;
  struct timespec next_period;

  period = (long) task_period;
  clock_gettime(CLOCK_MONOTONIC, &next_period);

  while (1) {
    timespec_add_ns(&next_period, period);
    clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &next_period, NULL);

    for (j=0; j<period/800; j++);

    pthread_mutex_lock(&mutex);

      for (j=0; j<period/600; j++);

    pthread_mutex_unlock(&mutex);

    for (j=0; j<period/2000; j++);

  }

  return (void *) 0;
}







void *treader(void *task_period) {
  int             error;
  int             unavailable;
  trace_event_id_t evid;
  long            period;
  struct timespec next_period;
  metrics_measurement_t  result;
  char                   name [METRICS_NAME_MAX];
  metrics_metricset_t    set;


  period = (long) task_period;
  clock_gettime(CLOCK_MONOTONIC, &next_period);

  rtl_printf("(reader) Starting...\n");
  rtl_printf("(reader) trid = %d\n", (int) trid);

  // Print out some information about the trace stream:
  posix_trace_eventtypelist_rewind(trid);
  posix_trace_eventtypelist_getnext_id (trid, &evid, &unavailable);
  rtl_printf("(reader) List of event types:\n");
  while (! unavailable) {
    posix_trace_eventid_get_name (trid, evid, str);
    rtl_printf("(reader)     Event %d name %s\n", (int) evid, str);
    posix_trace_eventtypelist_getnext_id (trid, &evid, &unavailable);
  }



  // Metrics initialization
  rtl_printf("(reader) Calling fill\n");
  metrics_metricset_fill (&set);
  rtl_printf("(reader) Calling init\n");
  metrics_init (trid, &set);
  rtl_printf("(reader) After init\n");

  // Now start reading events and printing them out:
  rtl_printf("(reader) MEASUREMENTS:\n");

  unavailable = 0;
  error = metrics_getnext_measurement (&result, &unavailable);

  while (! error && ! unavailable) {

    metrics_metric_getname (result.metric_id, name);
    
    rtl_printf ( "<MEASUREMENT>\n");
    rtl_printf ( "  <METRICID>%d</METRICID>\n",     result.metric_id);
    rtl_printf ( "  <METRICNAME>%s</METRICNAME>\n", name);
    rtl_printf ( "  <THREAD>%d</THREAD>\n",           result.thread_id);
    rtl_printf ( "  <OBJECT>%d</OBJECT>\n",           result.object_id);
    rtl_printf ( "  <DURATION>%ld.%09ld</DURATION>\n",  result.duration.tv_sec, result.duration.tv_nsec);
    rtl_printf ( "  <BEGIN>%ld.%09ld</BEGIN>\n",        result.begin.tv_sec, result.begin.tv_nsec);
    rtl_printf ( "  <END>%ld.%09ld</END>\n",            result.end.tv_sec, result.end.tv_nsec);
    rtl_printf ( "  <EVCOUNT>%d</EVCOUNT>\n",       result.events_count);
    rtl_printf ( "  <ID>%d</ID>\n",                result.id);
    rtl_printf ( "</MEASUREMENT>\n\n");

    
    timespec_add_ns(&next_period, period);
    clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &next_period, NULL);
    error = metrics_getnext_measurement (&result, &unavailable); 
  }

  rtl_printf ("(reader) Unavailable = %d, error = %d\n", unavailable, error);
  rtl_printf ("(reader) Exiting\n");

  return (void *) 0;
}




 


int init_module(void) {

  trace_attr_t   attr;
  pthread_attr_t thattr;
  int            error;
  struct sched_param   s_param;


  // Create stream attributes and set some of them:
  error = posix_trace_attr_init(&attr);
  error = posix_trace_attr_setstreamfullpolicy (&attr, POSIX_TRACE_UNTIL_FULL);
  error = posix_trace_attr_setmaxdatasize(&attr, 64);
  error = posix_trace_attr_setstreamsize(&attr, 40960);
  
 // Create the stream:
  error = posix_trace_create(0, &attr, &trid); 
  if (error) {
    return -1;
  }
  rtl_printf("(init_module) Stream %d created\n", (int) trid);
  // Start tracing:
  error = posix_trace_start(trid);
  if (error) {
    return -1;
  }

#ifdef _DEBUG
  rtl_printf("(init_module): Creating tasks\n");
#endif
  pthread_attr_init (&thattr); 

  // Create three tasks which just consume CPU:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 15;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task1, &thattr, just_execute, (void *) 100000000); 

  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 10;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task2, &thattr, just_execute, (void *) 200000000); 

  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 5;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task3, &thattr, just_execute, (void *) 200000000); 


  // Now create the "reader" task, which will online retrieve
  // the events from the trace stream:
  // (this task executes at the lowest priority of all)
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO);
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task_reader, &thattr, treader, (void *) 500000000); 

  return 0;
}





void cleanup_module(void) {

#ifdef _DEBUG
  rtl_printf("rtl_tasks: CLEANUP!!!\n");
#endif

  // Stop and shutdown the stream:
  posix_trace_shutdown(trid); 

  // Stop the tasks:
  pthread_cancel(task1);
  pthread_cancel(task2);
  pthread_cancel(task3);
  pthread_cancel(task_reader);

  pthread_join(task1,NULL); 
  pthread_join(task2,NULL); 
  pthread_join(task3,NULL); 
  pthread_join(task_reader,NULL); 

}
