/*
   FILE:
						       	
	rtl_tasks.c
								       	
   DESCRIPTION:							       	
	RT-Linux side of a sample application using the POSIX trace
	support.

	See README file for details.
					       				
   Copyright (C) 2001
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Ignacio Paches Giner
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        September 2001

*/

#include <rtl.h>
#include <time.h>
#include <pthread.h>
#include <rtl_sched.h>
#include <trace.h>
#include <rtl_fifo.h>
#include <unistd.h>
#include "./common.h"

#define _DEBUG

static int delay = 5;
static int fifo = 10;

MODULE_PARM (delay, "i");
MODULE_PARM (fifo, "i");

static trace_id_t trid;
static pthread_t /*writer_task, flush_task,*/ shutdown_task;
static pthread_t task1, task2, task3, task4;
static int fifo_desc;

typedef struct thread_data {
	struct timespec first_release;
	struct timespec period;
} thread_data;


void *
just_execute (void *data)
{
  struct thread_data *thr_data;
  struct timespec period;
  struct timespec next_period;
  struct timespec now;

  thr_data = (thread_data *) data;
  next_period   = thr_data->first_release;
  period        = thr_data->period;

  clock_gettime (CLOCK_MONOTONIC, &now);

  rtl_printf("(task1) now   = (%ld,%ld)\n",
		now.tv_sec, now.tv_nsec);

  rtl_printf("(task1) first = (%ld,%ld)  period = (%ld,%ld)\n",
		next_period.tv_sec, next_period.tv_nsec,
		period.tv_sec, period.tv_nsec);


  clock_nanosleep (CLOCK_MONOTONIC, TIMER_ABSTIME, &next_period, NULL);
  
  while (1)
    {
      timespec_add (&next_period, &period);
      clock_nanosleep (CLOCK_MONOTONIC, TIMER_ABSTIME, &next_period, NULL);
    }

  pthread_exit (0);

}


void *
task_to_flush (void *task_period)
{
  struct posix_trace_status_info status;
  struct timespec next_period;
  long period;
  int error;

  period = (long) task_period;

  clock_gettime (CLOCK_MONOTONIC, &next_period);
  timespec_add_ns (&next_period, period);

  while (1)
    {
      clock_nanosleep (CLOCK_MONOTONIC, TIMER_ABSTIME, &next_period, NULL);
      timespec_add_ns (&next_period, period);

      error = posix_trace_get_status (trid, &status);
      if (!error && status.posix_stream_status == POSIX_TRACE_RUNNING)
	{
#ifdef _DEBUG
	  rtl_printf ("  (flush_task): Flushing stream %d to log\n", trid);
#endif
	  posix_trace_flush (trid);
	}
    }

  return (void *) 0;
}






void *
task_to_shutdown (void *dummy)
{
  struct timespec wakeup_time;
  struct timespec retry_time;
  struct posix_trace_status_info status;

  clock_gettime (CLOCK_MONOTONIC, &wakeup_time);
  wakeup_time.tv_sec += delay;
  clock_nanosleep (CLOCK_MONOTONIC, TIMER_ABSTIME, &wakeup_time, NULL);

#ifdef _DEBUG
  rtl_printf ("  (shutdown_task): Waking up to shutdown the stream %d\n",
	      trid);
#endif
  posix_trace_get_status (trid, &status);
  retry_time = wakeup_time;
  while (status.posix_stream_flush_status == POSIX_TRACE_FLUSHING)
    {
      retry_time.tv_nsec += 100000000;
#ifdef _DEBUG
      rtl_printf
	("  (shutdown_task): There is a flush in progress. Sleeping...\n");
#endif
      clock_nanosleep (CLOCK_MONOTONIC, TIMER_ABSTIME, &retry_time, NULL);
      posix_trace_get_status (trid, &status);
    }

#ifdef _DEBUG
  rtl_printf ("  (shutdown_task): Shutting stream %d down\n", trid);
#endif
  posix_trace_shutdown (trid);

  return (void *) 0;
}






int
init_module (void)
{

  trace_attr_t attr;
  trace_event_set_t set;
  int error;
  char fifo_name[16];
  trace_event_id_t ev_info;
  struct idle_measurement_data info_data;
  struct thread_data thr_data;

  // Create and set the trace attribute:
  error = posix_trace_attr_init (&attr);
  error =
    posix_trace_attr_setstreamfullpolicy (&attr, POSIX_TRACE_UNTIL_FULL);
  error = posix_trace_attr_setmaxdatasize (&attr, 32);
  error = posix_trace_attr_setstreamsize (&attr, 8096000);


  // Open the fifo
  sprintf (fifo_name, "/dev/rtf%d", fifo);
#ifdef _DEBUG
  rtl_printf ("(init_module) Creating fifo No. %d (%s)\n", fifo, fifo_name);
#endif
  rtf_create (fifo, 100000);

#ifdef _DEBUG
  rtl_printf ("(init_module) Opening fifo No. %d (%s)\n", fifo, fifo_name);
#endif

  fifo_desc = open (fifo_name, O_NONBLOCK);
  if (fifo_desc < 0)
    {
      rtl_printf ("(init_module): Cannot open %s for writing\n", fifo_name);
      return -1;
    }

  // Create the stream:
#ifdef _DEBUG
  rtl_printf ("(init_module) Creating stream witlog %d\n", fifo_desc);
#endif
  error = posix_trace_create_withlog (0, &attr, fifo_desc, &trid);
  if (error)
    {
      rtl_printf ("(init_module): Cannot create trace with log %d\n",
		  fifo_desc);
      return -1;
    }


  // Set the stream filter to only record user events:
#ifdef _DEBUG
  rtl_printf ("(init_module) Calling to filter system events\n");
#endif
  
  posix_trace_eventset_fill(&set, POSIX_TRACE_SYSTEM_EVENTS);
  posix_trace_eventset_del(POSIX_TRACE_CSWITCH_TOLINUX, &set);
  posix_trace_eventset_del(POSIX_TRACE_CSWITCH_TOTASK, &set);
  //error = posix_trace_set_filter(trid, 
  //  				(const trace_event_set_t *) &set, 
  //				POSIX_TRACE_SET_EVENTSET);
   

  // Start tracing:
#ifdef _DEBUG
  rtl_printf ("(init_module) Starting trace stream %d\n", (int) trid);
#endif
  error = posix_trace_start (trid);
  if (error)
    {
      rtl_printf ("(init_module): Cannot start trace %d\n", trid);
      return -1;
    }

  posix_trace_eventid_open ("idle-measurement-info", &ev_info);
  clock_gettime (CLOCK_MONOTONIC, &info_data.first_release);
  rtl_printf("(init_module) now           = (%ld,%ld)\n",
		info_data.first_release.tv_sec, 
		info_data.first_release.tv_nsec);



  timespec_add_ns (&info_data.first_release, (long long) 100000000L);
  rtl_printf("(init_module) first_release = (%ld,%ld)\n",
		info_data.first_release.tv_sec, 
		info_data.first_release.tv_nsec);

  info_data.hyperperiod.tv_sec = 0L;
  info_data.hyperperiod.tv_nsec = 200000000L;

  posix_trace_event (ev_info, &info_data, sizeof (info_data));

#ifdef _DEBUG
  rtl_printf ("(init_module) Creating tasks\n");
#endif

  thr_data.first_release  = info_data.first_release;
  thr_data.period.tv_sec  = 0L;
  thr_data.period.tv_nsec = 10000000L;
  pthread_create (&task1, NULL, just_execute, (void *) &thr_data);
  
  thr_data.period.tv_sec  = 0L;
  thr_data.period.tv_nsec = 20000000L;
  pthread_create (&task2, NULL, just_execute, (void *) &thr_data);

  thr_data.period.tv_sec  = 0L;
  thr_data.period.tv_nsec = 25000000L;
  pthread_create (&task3, NULL, just_execute, (void *) &thr_data);

  thr_data.period.tv_sec  = 0L;
  thr_data.period.tv_nsec = 40000000L;
  pthread_create (&task4, NULL, just_execute, (void *) &thr_data);

  //pthread_create (&flush_task, NULL, task_to_flush, (void *) 500000000); 
  pthread_create (&shutdown_task, NULL, task_to_shutdown, NULL);

  return 0;
}



void
cleanup_module (void)
{

#ifdef _DEBUG
  rtl_printf ("rtl_tasks: CLEANUP!!!\n");
#endif

  pthread_cancel (task1);
  pthread_cancel (task2);
  pthread_cancel (task3);
  pthread_cancel (task4);
  //pthread_cancel (flush_task);
  pthread_cancel (shutdown_task);

  pthread_join (task1, NULL);
  pthread_join (task2, NULL);
  pthread_join (task3, NULL);
  pthread_join (task4, NULL);
  //pthread_join (flush_task, NULL);
  pthread_join (shutdown_task, NULL);

  // Close the fifo 
  close (fifo_desc);
  rtf_destroy (fifo);

}
