/*
   FILE:							       	
	linux_analyzer.c
								       	
   DESCRIPTION:						
	      	
	Linux side of a sample application using the POSIX trace
	support. This Linux program retrieves all the information (and
	the user trace events) from the log stream created by the
	RT-Linux part (rtl_tasks.c).

	See README file for details.
					       				
   Copyright (C) 2001
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Ignacio Paches Giner
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        December 2003

*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>



#include <trace.h>


#define MAX_DATA_SIZE 16


int main(int argc, char *argv[]) {

  char         log_filename[256];
  char         data[MAX_DATA_SIZE];
  char         str[TRACE_NAME_MAX];
  size_t       data_len;
  size_t       size;
  int          unavailable;
  int          pol;
  int          error;
  trace_event_id_t ev;
  trace_attr_t trace_attr;
  trace_id_t   trace_id;
  struct posix_trace_event_info event;
  struct posix_trace_status_info status;
  struct timespec  time;
  int   *i;
  char  *c;
  struct posix_trace_sysevent_data *sysdata;

  int   log_fd;
  int   nevents = 0;
  
  if (argc < 2) {
    fprintf(stderr, "Use: linux_analyzer logfile\n");
    exit(1);
  } else if (argc == 2) {
    strncpy(log_filename, argv[1], sizeof(log_filename));
  }


  if ((log_fd = open(log_filename, O_RDONLY)) < 0) {
    fprintf(stderr, "linux_analyzer: Error while opening log file %s\n", 
	    log_filename);
    exit(1);
  }

  error = posix_trace_open(log_fd, &trace_id);
  if (error < 0) {
    fprintf(stderr, 
	    "linux_analyzer: Error while opening pre-recorded stream\n");
    exit(1);

  }

  // Retrieve the attributes of this stream and print all the info:
  error = posix_trace_get_attr(trace_id, &trace_attr);
  fprintf(stdout, "get attr (%d)\n", error);
  
  error = posix_trace_attr_getgenversion(&trace_attr, str);
  fprintf(stdout, "get genversion (%d): %s\n", error, str);

  error = posix_trace_attr_getname(&trace_attr, str);
  fprintf(stdout, "get name (%d): %s\n", error, str);

  error = posix_trace_attr_getcreatetime(&trace_attr, &time);
  fprintf(stdout, "get create time (%d): %ld.%ld\n", error, 
	  time.tv_sec, time.tv_nsec);

  error = posix_trace_attr_getclockres(&trace_attr, &time);
  fprintf(stdout, "get clock res (%d): %ld.%ld\n", error, 
	  time.tv_sec, time.tv_nsec);

  error = posix_trace_attr_getstreamfullpolicy(&trace_attr, &pol);
  fprintf(stdout, "get full policy (%d): %d\n", error, pol);

  error = posix_trace_attr_getmaxusereventsize(&trace_attr, 0, &size);
  fprintf(stdout, "get max user event size (%d): %d\n", error, size);

  error = posix_trace_attr_getmaxsystemeventsize(&trace_attr, &size);
  fprintf(stdout, "get max system event size (%d): %d\n", error, size);

  error = posix_trace_attr_getmaxdatasize(&trace_attr, &size);
  fprintf(stdout, "get max data size (%d): %d\n", error, size);

  error = posix_trace_attr_getstreamsize(&trace_attr, &size);
  fprintf(stdout, "get stream size (%d): %d\n", error, size);


  // -------------------------------------------------------------------------
  // Print the full list of event types available in the stream:

  error = posix_trace_eventtypelist_rewind(trace_id);
  fprintf(stdout, "event list rewind (%d)\n", error);

  error = posix_trace_eventtypelist_getnext_id (trace_id, &ev, &unavailable);
  while( ! unavailable && ! error) {
    
    error = posix_trace_eventid_get_name (trace_id, ev, str);

    if (! error) {
      fprintf(stdout, "event getname (%d): %s\n", error, str);
    }
    // Get the next one:
    error = posix_trace_eventtypelist_getnext_id (trace_id, &ev, 
						  &unavailable);
  }


  // -------------------------------------------------------------------------
  // Print the current status of the stream :
  error = posix_trace_get_status(trace_id, &status);
  fprintf(stdout, "event get status (%d): \n", error);

  fprintf(stdout, "\tstream status         : ");
  if (status.posix_stream_status == POSIX_TRACE_RUNNING) {
	 fprintf(stdout, "POSIX_TRACE_RUNNING\n");
  } else if (status.posix_stream_status == POSIX_TRACE_SUSPENDED) {
	 fprintf(stdout, "POSIX_TRACE_SUSPENDED\n");
  } else {
	 fprintf(stdout, " # UNKNOWN #\n");
  }

  fprintf(stdout, "\tstream full status    : ");
  if (status.posix_stream_full_status == POSIX_TRACE_FULL) {
	 fprintf(stdout, "POSIX_TRACE_FULL\n");
  } else if (status.posix_stream_full_status == POSIX_TRACE_NOT_FULL) {
	 fprintf(stdout, "POSIX_TRACE_NOT_FULL\n");
  } else {
	 fprintf(stdout, " # UNKNOWN #\n");
  }

  fprintf(stdout, "\tstream overrun status : ");
  if (status.posix_stream_overrun_status == POSIX_TRACE_OVERRUN) {
	 fprintf(stdout, "POSIX_TRACE_OVERRUN\n");
  } else if (status.posix_stream_overrun_status == POSIX_TRACE_NO_OVERRUN) {
	 fprintf(stdout, "POSIX_TRACE_NO_OVERRUN\n");
  } else {
	 fprintf(stdout, " # UNKNOWN #\n");
  }

  fprintf(stdout, "\tstream flush status   : ");
  if (status.posix_stream_flush_status == POSIX_TRACE_FLUSHING) {
	 fprintf(stdout, "POSIX_TRACE_FLUSHING\n");
  } else if (status.posix_stream_flush_status == POSIX_TRACE_NOT_FLUSHING) {
	 fprintf(stdout, "POSIX_TRACE_NOT_FLUSHING\n");
  } else {
	 fprintf(stdout, " # UNKNOWN #\n");
  }

  fprintf(stdout, "\tstream flush error    : %d\n",
	  status.posix_stream_flush_error);

  fprintf(stdout, "\tlog overrun status    : ");
  if (status.posix_log_overrun_status == POSIX_TRACE_OVERRUN) {
	 fprintf(stdout, "POSIX_TRACE_OVERRUN\n");
  } else if (status.posix_log_overrun_status == POSIX_TRACE_NO_OVERRUN) {
	 fprintf(stdout, "POSIX_TRACE_NO_OVERRUN\n");
  } else {
	 fprintf(stdout, " # UNKNOWN #\n");
  }

  fprintf(stdout, "\tlog full status       : ");
  if (status.posix_log_overrun_status == POSIX_TRACE_FULL) {
	 fprintf(stdout, "POSIX_TRACE_FULL\n");
  } else if (status.posix_log_overrun_status == POSIX_TRACE_NOT_FULL) {
	 fprintf(stdout, "POSIX_TRACE_NOT_FULL\n");
  } else {
	 fprintf(stdout, " # UNKNOWN #\n");
  }



  
  // -------------------------------------------------------------------------
  // Start retrieving events while the stream is open:
  fprintf(stdout, "\n\nNow retrieving events: \n");


  while(1) {

    error = posix_trace_getnext_event(trace_id, &event, &data, sizeof(data),
				      &data_len, &unavailable);

    if(error) {
      fprintf(stdout,"No more events (%d). Exiting\n", error);
      exit(0);

    } else if (unavailable) {
      fprintf(stdout, "   Event unavailable\n");
      exit(0);

    } else {

      fprintf(stdout, " ** New event --( %2d )------------------------\n", ++nevents);
      fprintf(stdout, " Event identifier = %d\n", event.posix_event_id);
      fprintf(stdout, " Process ID       = %d\n", event.posix_pid);
      fprintf(stdout, " Address          = %x\n", (int) event.posix_prog_address);
      fprintf(stdout, " Truncation status= %d\n", event.posix_truncation_status);
      fprintf(stdout, " Timestamp        = %ld.%09ld\n", event.posix_timestamp.tv_sec, 
	      event.posix_timestamp.tv_nsec);
      fprintf(stdout, " Thread ID        = %d\n",   (int) event.posix_thread_id);
      
      posix_trace_eventid_get_name (trace_id, event.posix_event_id, str);
      fprintf(stdout, " Event Name       = %s\n", str);

      // Now switch depending on the event type (name):
      if (!strcmp(str,"user event char")) {
	c = (char *) data;
	fprintf(stdout, " Data (char)      = %c\n",  *c);
      } 

      else if (!strcmp(str,"user event int")) {
	i = (int *) data;
	fprintf(stdout, " Data (int)       = %d\n",  *i);
      }

      else if (!strcmp(str,"user event string")) {
	fprintf(stdout, " Data (string)    = %s\n",  (char *) data);
	
      } 

      else if (!strcmp(str,"user event linux")) {
	i = (int *) data;
	fprintf(stdout, " Data (int)       = %d\n",  *i);
      } 

      /* ----------------------------------------------------------- */
      // Eventos de Agus:
      //
      else if (event.posix_event_id >=POSIX_TRACE_SYSCALL &&
	       event.posix_event_id <= POSIX_TRACE_SCHED_START) {
	sysdata = (posix_trace_sysevent_data *) data;
	i = (int *) data;
	fprintf(stdout, " Data (system event)\n");
	fprintf(stdout, "    data.index       = %d\n",  sysdata->index);
	fprintf(stdout, "    data.task        = %d\n",  sysdata->task);
      } 
      /* ----------------------------------------------------------- */

      else {
	fprintf(stdout, " Data unknown\n" );
      } 

      fprintf(stdout, " Data length      = %d\n",  data_len);
      fprintf(stdout, " -----------------------------------------------\n\n");
    }

    strncpy(data, "\0", 1);
    
  }

  return 0;


}
