/*
   FILE:							       	
	trace.h							       
								       	
   DESCRIPTION:							       	
        Public POSIX tracing definitions 
					       				
   Copyright (C) 2001,2002
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Ignacio Paches Giner
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        April 2002

*/

#ifndef __POSIX_TRACE_H
#define __POSIX_TRACE_H



// Import which POSIX tracing options are currently supported:
#include "trace/trace_options.h" 


#if defined _POSIX_TRACE_CORE || defined _POSIX_TRACE_LOGRET


// ---
// BEGIN include section
#ifdef MODULE
#include <linux/time.h>
#include <rtl_sched.h>
#include <rtl_mutex.h>
#else
#include <sys/time.h>
#include <pthread.h>
#include <sys/types.h>
#endif

#include <rtl_time.h>



#include "trace/trace.h" // POSIX tracing private definitions
// END include section
// ---



/* ---------------------------------------------------------------------
   GENERAL NOTE:
      This header file is (mostly) ordered and labelled according to the 
      POSIX "P1003.1q: Tracing" standard document.
   ---------------------------------------------------------------------
*/



/* 
   -- 2.5 PRIMITIVE SYSTEM DATA TYPES ----------------------------------
*/

// Used to define a trace stream attributes object:
typedef  _trace_attr_t       trace_attr_t;

// Used to identify a trace stream:  
typedef  _trace_id_t         trace_id_t;

// Used to identify a trace event type:
typedef  _trace_event_id_t   trace_event_id_t;

// Used to define a set of trace event type identifiers:
typedef  _trace_event_set_t  trace_event_set_t;



/* 
   -- 2.8.4. RUN-TIME INVARIANT VALUES ---------------------------------
*/
#define TRACE_SYS_MAX             _POSIX_TRACE_SYS_MAX
#define TRACE_USER_EVENT_MAX      _POSIX_TRACE_USER_EVENT_MAX 
#define TRACE_NAME_MAX            _POSIX_TRACE_NAME_MAX
#define TRACE_EVENT_NAME_MAX      _POSIX_TRACE_EVENT_NAME_MAX 




/* 
   -- 24.2.1. DATA DEFINITIONS: STRUCTURES -----------------------------
*/

// 24.2.1.1. posix_trace_status_info

typedef struct posix_trace_status_info {
  int posix_stream_status;         // Operating mode of the trace stream
  int posix_stream_full_status;    // Full status of the trace stream
  int posix_stream_overrun_status; // Trace event were lost?
#if defined _POSIX_TRACE_TOLOG || defined _POSIX_TRACE_LOGRET
  int posix_stream_flush_status;   // Is there a flush in progress?
  int posix_stream_flush_error;    // Error during last flush operation?
  int posix_log_overrun_status;    // Trace events lost in the log?
  int posix_log_full_status;       // Full status of the trace log
#endif
} posix_trace_status_info;


// Pre-defined values for some of the fields:

enum /* posix_stream_status */ { 
  POSIX_TRACE_RUNNING,
  POSIX_TRACE_SUSPENDED
};

enum /* posix_stream_full_status and posix_log_full_status */ {
  POSIX_TRACE_FULL,
  POSIX_TRACE_NOT_FULL
};

enum /* posix_stream_overrun_status and posix_log_overrun_status */ {
  POSIX_TRACE_OVERRUN,
  POSIX_TRACE_NO_OVERRUN
};

#if defined _POSIX_TRACE_TOLOG || defined _POSIX_TRACE_LOGRET
enum /* posix_stream_flush_status */ {
  POSIX_TRACE_FLUSHING,
  POSIX_TRACE_NOT_FLUSHING
};
#endif



// 24.2.1.2. posix_trace_event_info


typedef struct posix_trace_event_info {
  trace_event_id_t 
            posix_event_id;        // Trace event type idenfication
  pid_t     posix_pid;             // Id of the event generator process
  void    * posix_prog_address;    // Address at which trace point was invoked
  int       posix_truncation_status;// Truncation info on this event
  pthread_t posix_thread_id;       // Id of the event generator thread
  struct timespec 
            posix_timestamp;       // Event generation time
} posix_trace_event_info;


// Pre-defined values for some of the fields:

enum /* posix_truncation_status */ {
  POSIX_TRACE_NOT_TRUNCATED,
  POSIX_TRACE_TRUNCATED_RECORD,
  POSIX_TRACE_TRUNCATED_READ,
};





/*
   -- 24.2.2. TRACE STREAM ATTRIBUTES --------------------------------------
*/


// Pre-defined values for some of the attributes of trace_attr_t:

// Common values for stream's and log's full policies:
#define POSIX_TRACE_LOOP           1
#define POSIX_TRACE_UNTIL_FULL     2

#ifdef _POSIX_TRACE_TOLOG
// third value for stream full policy:
#define POSIX_TRACE_FLUSH          3
// third value for log full policy:
#define POSIX_TRACE_APPEND         3
#endif



#ifdef _POSIX_TRACE_INHERITANCE
// Values for inheritance policy:
#define POSIX_TRACE_INHERITED       1
#define POSIX_TRACE_CLOSE_FOR_CHILD 2
#endif





/*
   -- 24.2.3. TRACE EVENT TYPE DEFINITIONS ---------------------------------
*/

// NOTE: User events are numbered from 0 to TRACE_USER_EVENT_MAX,
//       where the last one is UNNAMED_USEREVENT. This leaves exactly
//       TRACE_USER_EVENT_MAX number of events available for the application.
//       System events are numbered starting on TRACE_USER_EVENT_MAX+1


// 24.2.3.1. System events

#define TRACE_FIRST_SYSTEM_EVENT   (TRACE_USER_EVENT_MAX+1)

// 1) posix_trace_error
#define POSIX_TRACE_ERROR          TRACE_FIRST_SYSTEM_EVENT

typedef struct posix_trace_error_event_data {
  int   error;
} posix_trace_error_event_data;

// 2) posix_trace_start
#define POSIX_TRACE_START          (TRACE_FIRST_SYSTEM_EVENT+1)

#ifdef _POSIX_TRACE_EVENT_FILTER
typedef struct posix_trace_start_event_data {
  trace_event_set_t  event_filter;
} posix_trace_start_event_data;
#endif

// 3) posix_trace_stop
#define POSIX_TRACE_STOP           (TRACE_FIRST_SYSTEM_EVENT+2)

typedef struct posix_trace_stop_event_data {
  int   automatic;
} posix_trace_stop_event_data;

// 4) posix_trace_overflow
#define POSIX_TRACE_OVERFLOW       (TRACE_FIRST_SYSTEM_EVENT+3)

// 5) posix_trace_resume
#define POSIX_TRACE_RESUME         (TRACE_FIRST_SYSTEM_EVENT+4)  

// 6) posix_trace_filter
// NOTE: This event is defined but not traced if
// _POSIX_TRACE_EVENT_FILTER support is not available
#define POSIX_TRACE_FILTER         (TRACE_FIRST_SYSTEM_EVENT+5)  

typedef struct posix_trace_filter_event_data {
  trace_event_set_t  old_event_filter;
  trace_event_set_t  new_event_filter;
} posix_trace_filter_event_data;



// 24.2.3.2. User events

// 1) posix_trace_unnamed_userevent 
#define POSIX_TRACE_UNNAMED_USEREVENT  TRACE_USER_EVENT_MAX





/*
   -- 24.3. TRACE FUNCTIONS ------------------------------------------------
*/

// 24.3.2. Attribute initialization and destruction
#ifdef _POSIX_TRACE_CORE
extern int posix_trace_attr_init (trace_attr_t * attr);

extern int posix_trace_attr_destroy (trace_attr_t *attr);
#endif

// 24.3.3. Retrieve and set information about a trace stream

extern int posix_trace_attr_getgenversion (const trace_attr_t *attr, 
					   char *genversion);

extern int posix_trace_attr_getname (const trace_attr_t *attr,
				     char *tracename);

extern int posix_trace_attr_setname (trace_attr_t *attr, 
				     const char *tracename);

extern int posix_trace_attr_getcreatetime (const trace_attr_t *attr, 
					   struct timespec *createtime);

extern int posix_trace_attr_getclockres (const trace_attr_t *attr,
					 struct timespec * resolution);


// 24.3.4. Retrieve and set the behavior of a trace stream

#ifdef _POSIX_TRACE_INHERITANCE
extern int posix_trace_attr_getinherited (const trace_attr_t *attr, 
					  int *inheritancepolicy);

#ifdef _POSIX_TRACE_CORE
extern int posix_trace_attr_setinherited (trace_attr_t *attr, 
					  int inheritancepolicy);
#endif
#endif

extern int posix_trace_attr_getstreamfullpolicy (const trace_attr_t *attr, 
						 int *streampolicy);

#ifdef _POSIX_TRACE_CORE
extern int posix_trace_attr_setstreamfullpolicy (trace_attr_t *attr, 
						 int streampolicy);
#endif


#ifdef _POSIX_TRACE_TOLOG
extern int posix_trace_attr_getlogfullpolicy (const trace_attr_t *attr, 
					      int *logpolicy);

#ifdef _POSIX_TRACE_CORE
extern int posix_trace_attr_setlogfullpolicy (trace_attr_t *attr, 
					      int logpolicy);
#endif
#endif


// 24.3.5. Retrieve and set trace stream size attributes

extern int posix_trace_attr_getmaxusereventsize (const trace_attr_t *attr, 
						 size_t data_len, 
						 size_t *eventsize);

extern int posix_trace_attr_getmaxsystemeventsize (const trace_attr_t *attr,
						   size_t *eventsize);

extern int posix_trace_attr_getmaxdatasize (const trace_attr_t *attr, 
					    size_t *maxdatasize);

extern int posix_trace_attr_getstreamsize (const trace_attr_t *attr, 
					   size_t *streamsize);

#ifdef _POSIX_TRACE_CORE
extern int posix_trace_attr_setmaxdatasize (trace_attr_t *attr, 
					    size_t maxdatasize);

extern int posix_trace_attr_setstreamsize (trace_attr_t *attr, 
					   size_t streamsize);

#endif

#if defined _POSIX_TRACE_TOLOG || defined _POSIX_TRACE_LOGRET
extern int posix_trace_attr_getlogsize (const trace_attr_t *attr, 
					size_t *logsize);
#endif

#ifdef _POSIX_TRACE_TOLOG
extern int posix_trace_attr_setlogsize(trace_attr_t *attr, 
				       size_t logsize);
#endif


// 24.3.6. Trace stream initialization, flush and shutdown

#ifdef _POSIX_TRACE_ONLINE
extern int posix_trace_create (pid_t pid,
			       const trace_attr_t *attr,
			       trace_id_t *trid);
#endif

#ifdef _POSIX_TRACE_TOLOG
extern int posix_trace_create_withlog (pid_t pid,
				       const trace_attr_t *attr,
				       int file_desc,
				       trace_id_t *trid);

extern int posix_trace_flush (trace_id_t trid);
#endif

#ifdef _POSIX_TRACE_CORE
extern int posix_trace_shutdown (trace_id_t trid);


// 24.3.7. Clear trace stream and trace log

extern int posix_trace_clear (trace_id_t trid);
#endif

// 24.3.8. Manipulate trace event type identifier

#ifdef _POSIX_TRACE_EVENT_FILTER
extern int posix_trace_trid_eventid_open (trace_id_t trid,
					  const char *event_name,
					  trace_event_id_t *event);
#endif 

extern int posix_trace_eventid_get_name (trace_id_t trid,
					 trace_event_id_t event,
					 char *event_name);

extern int posix_trace_eventid_equal (trace_id_t trid,
				      trace_event_id_t event1,
				      trace_event_id_t event2);


// 24.3.9. Iterate over a mapping of trace event types

extern int posix_trace_eventtypelist_getnext_id (trace_id_t trid,
						 trace_event_id_t *event,
						 int *unavailable);

extern int posix_trace_eventtypelist_rewind (trace_id_t trid);


// 24.3.10. Manipulate trace event type sets

#ifdef _POSIX_TRACE_EVENT_FILTER


// Argument 'what' in posix_trace_eventset_fill() can be:
#define  POSIX_TRACE_WOPID_EVENTS    1
#define  POSIX_TRACE_SYSTEM_EVENTS   2
#define  POSIX_TRACE_ALL_EVENTS      3

extern int posix_trace_eventset_empty (trace_event_set_t *set);

extern int posix_trace_eventset_fill (trace_event_set_t *set,
				      int what);

extern int posix_trace_eventset_add (trace_event_id_t event_id,
				     trace_event_set_t *set);

extern int posix_trace_eventset_del (trace_event_id_t event_id,
				     trace_event_set_t *set);

//extern int posix_trace_eventset_complement(trace_event_set_t *set);

extern int posix_trace_eventset_ismember (trace_event_id_t event_id,
					  const trace_event_set_t *set,
					  int *ismember);


// 24.3.11. Manipulate trace event type sets

// Argument 'how' in posix_trace_set_filter() can be:
#define  POSIX_TRACE_SET_EVENTSET    1
#define  POSIX_TRACE_ADD_EVENTSET    2
#define  POSIX_TRACE_SUB_EVENTSET    3

extern int posix_trace_set_filter (trace_id_t trid, 
				   const trace_event_set_t *set,
				   int how);


extern int posix_trace_get_filter (trace_id_t trid, 
				   trace_event_set_t *set);
#endif


// 24.3.12. Trace start and stop
#ifdef _POSIX_TRACE_CORE
extern int posix_trace_start (trace_id_t trid);

extern int posix_trace_stop (trace_id_t trid);
#endif

// 24.3.13. Instrumenting application code
#ifdef _POSIX_TRACE_UEVENTS
extern int posix_trace_eventid_open (const char *event_name, 
				     trace_event_id_t *event_id);

extern void posix_trace_event (trace_event_id_t event_id, 
			       const void *data_ptr,      
			       size_t data_len);
#endif

// 24.3.14. Trace log management

#ifdef _POSIX_TRACE_LOGRET

extern int posix_trace_open (int file_desc,
			     trace_id_t *trid);

extern int posix_trace_rewind (trace_id_t trid);

extern int posix_trace_close (trace_id_t trid);

#endif


// 24.3.15. Retrieve the trace attributes or trace statuses
#if defined _POSIX_TRACE_ONLINE || defined _POSIX_TRACE_LOGRET
extern int posix_trace_get_attr (trace_id_t trid,
				 trace_attr_t *attr);

extern int posix_trace_get_status (trace_id_t trid,
				   struct posix_trace_status_info *statusinfo);


// 24.3.16. Retrieve a trace event

extern int posix_trace_getnext_event (trace_id_t trid,
				      struct posix_trace_event_info *event, 
				      void *data,
				      size_t num_bytes,
				      size_t *data_len,
				      int *unavailable);
#endif

#ifdef _POSIX_TRACE_ONLINE
extern int posix_trace_timedgetnext_event (trace_id_t trid,
					   struct posix_trace_event_info 
					                             *event,
					   void *data,
					   size_t num_bytes,
					   size_t *data_len,
					   int *unavailable,
					   const struct timespec *abs_timeout);

extern int posix_trace_trygetnext_event (trace_id_t trid,
					 struct posix_trace_event_info *event,
					 void *data,
					 size_t num_bytes,
					 size_t *data_len,
					 int *unavailable);
#endif





/*
  ===========================================================================
  IMPLEMENTATION-DEPENDENT DEFINITIONS
  ===========================================================================
*/


/*
   NEW SYSTEM EVENT TYPES (DEFINED BY THE IMPLEMENTATION)
*/ 

//  Maximum values for system event types and total event types
#define  TRACE_SYSTEM_EVENT_MAX  _POSIX_TRACE_SYSTEM_EVENT_MAX  

#define  TRACE_EVENT_MAX         (TRACE_USER_EVENT_MAX + \
                                 TRACE_SYSTEM_EVENT_MAX)

// Data structure common to all system events:
typedef struct posix_trace_sysevent_data {
  int     index;   
  int     task;  // 0 = linux;  { 1, 2, 3,...} = App rt tasks
  int     priority;
  int     return_value; 
  union   other_data {
    struct mutex_data {
      int  mutex;
    } mutex_data;

    struct cond_data {
      int  mutex;
      int  cond;
    } cond_data;    

    struct timespec delay_until;
  } other_data;
} posix_trace_sysevent_data;


// Identifier definitions and index values:

// 1) Context switch:
#define POSIX_TRACE_CSWITCH        (TRACE_FIRST_SYSTEM_EVENT+6)


// 2) State change for tasks:
#define POSIX_TRACE_TASK_NEW_STATE (TRACE_FIRST_SYSTEM_EVENT+7)

#define TASK_NEW_STATE_READY       1
#define TASK_NEW_STATE_UNREADY     2
#define TASK_NEW_STATE_RELEASED    3


// 3) RT-Linux scheduler starts:
#define POSIX_TRACE_SCHED_START    (TRACE_FIRST_SYSTEM_EVENT+8)
// data.index = 1 <--> timer interrupt


// 4) RT-Linux scheduler ends:
#define POSIX_TRACE_SCHED_END      (TRACE_FIRST_SYSTEM_EVENT+9)


// 5) Start of processing a system call:
#define POSIX_TRACE_SYSCALL        (TRACE_FIRST_SYSTEM_EVENT+10)

#define SYSCALL_CLOCK_NANOSLEEP           1
#define SYSCALL_PTHREAD_CREATE            2
#define SYSCALL_SCHED_YIELD               3
#define SYSCALL_PTHREAD_KILL              4
#define SYSCALL_PTHREAD_EXIT              5
#define SYSCALL_PTHREAD_MUTEX_LOCK        6
#define SYSCALL_PTHREAD_MUTEX_UNLOCK      7
#define SYSCALL_PTHREAD_COND_WAIT         8
#define SYSCALL_PTHREAD_COND_SIGNAL       9
#define SYSCALL_PTHREAD_COND_BROADCAST    10
#define SYSCALL_PTHREAD_WAIT_NP           11
#define SYSCALL_PTHREAD_SUSPEND_NP        12
#define SYSCALL_PTHREAD_WAKEUP_NP         13
#define SYSCALL_PTHREAD_DELETE_NP         14


// 6) End of processing a system call:
#define POSIX_TRACE_SYSCALL_END    (TRACE_FIRST_SYSTEM_EVENT+11)


// 7) Mutex lock:
#define POSIX_TRACE_MUTEX_LOCK     (TRACE_FIRST_SYSTEM_EVENT+12) 

#define  MUTEX_LOCK_DENIED         1
#define  MUTEX_LOCK_GRANTED        2


// 8) Mutex unlock:
#define POSIX_TRACE_MUTEX_UNLOCK   (TRACE_FIRST_SYSTEM_EVENT+13)


// 9) Condition variable wait:
#define POSIX_TRACE_COND_WAIT      (TRACE_FIRST_SYSTEM_EVENT+14)


// 10) Condition variable broadcast:
#define POSIX_TRACE_COND_BROADCAST (TRACE_FIRST_SYSTEM_EVENT+15)


// 11) Context switch from a task to Linux
#define POSIX_TRACE_CSWITCH_TOLINUX (TRACE_FIRST_SYSTEM_EVENT+16)


// 12) Context switch from Linux to a task
#define POSIX_TRACE_CSWITCH_TOTASK  (TRACE_FIRST_SYSTEM_EVENT+17)


// Reset this if system type events are added or removed
#define TRACE_LAST_SYSTEM_EVENT    POSIX_TRACE_CSWITCH_TOTASK


/* 
   All the instrumentation points for system event types inside the
   RT-Linux scheduler are coded as macros (this makes the code easier
   to read and maintain). The next header file contains the macro
   definitions.
 */
#include "trace/instr_macros.h"




/* 
   RT-FIFO related definitions
*/
#define TRACE_TARGET_FIFO         _POSIX_TRACE_LINUX_FIFO



// NOTE: When using FIFOs, we need to bound the size of event data.
// If the event data size is greater than this value, the data will 
// be truncated to that size:
// (This only affects events which do pass through FIFOs from Linux
// target processes
 
#define MAX_FIFO_DATA_SIZE    256





#ifdef MODULE
/* 
   GLOBAL FILTER for SYSTEM EVENTS

   This is a global filter exported to other parts of RT-Linux 
   This filter contains the logical AND of the filters of all the
   running streams. Therefore, if a system event is filtered in the
   global filter, it is so in every active stream, and we don't spend
   any time in tracing it!
   
*/

extern int global_system_events_filter;


#define TRACE_SYSTEM_EVENT(ev)\
                       ( \
                        !(global_system_events_filter & 1<<(ev-32))  \
                       )


/* 
   FUNCTION for TRACING SYSTEM EVENTS

   Exported only to other RTL modules
*/

extern void posix_trace_system_event(trace_event_id_t event_id, 
				     const void *data_ptr,      
				     size_t data_len,  int calling_address);

#endif





#endif // #if defined _POSIX_TRACE_CORE || defined _POSIX_TRACE_LOGRET


#endif // #ifndef __POSIX_TRACE_H


