#ifdef MODULE
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/slab.h>
#endif


#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <trace.h>
#include <stdio.h>

#include <metrics.h>

#include <metrics/metrics_events.h>
#include <metrics/metrics_register.h>
#include <metrics/metrics_definitions.h>

///////////////////////////////////////////////
///////////////////////////////////////////////
//  AUTOMATA
///////////////////////////////////////////////
///////////////////////////////////////////////

mtri_automaton_instance_descriptor_t *automata_list_head_ptr;


// DEBUG STUFF

#ifndef MODULE
#ifdef MTRI_DEBUG

static int
print_state (int state) {

  switch (state) {
  case   START:  fprintf (stderr,"START") ;break;
  case   BEGIN:  fprintf (stderr,"BEGIN") ;break;
  case   IN:     fprintf (stderr,"IN")    ;break;
  case   OUT:    fprintf (stderr,"OUT")   ;break;
  case   CANCEL: fprintf (stderr,"CANCEL");break;
  case   END:    fprintf (stderr,"END")   ;break;
  }

  }

static int 
show_automaton (mtri_automaton_instance_descriptor_t *aut)

{
  fprintf (stderr, "AUTOMATON\n");
  fprintf (stderr, "  %20s = %d\n", "thread", aut->thread);
  fprintf (stderr, "  %20s = %d\n", "measuring", aut->measuring);
  fprintf (stderr, "  %20s = %d\n", "in_segment", aut->in_segment);
  fprintf (stderr, "  %20s = %9d.%09d\n", "begin",  aut->begin.tv_sec, aut->begin.tv_nsec);
  fprintf (stderr, "  %20s = %9d.%09d\n", "end",  aut->end.tv_sec, aut->end.tv_nsec);
  fprintf (stderr, "  %20s = %9d.%09d\n", "partial_first",  aut->partial_first.tv_sec, aut->partial_first.tv_nsec);
  fprintf (stderr, "  %20s = %9d.%09d\n", "sum",  aut->sum.tv_sec, aut->sum.tv_nsec);

  fprintf (stderr, " \n");
  
  return 0;
}
#endif // MTRI_DEBUG
#endif // MODULE

// END DEBUG STUFF




///////////////////////////////////////////////
//  mtri_automata_initialize
///////////////////////////////////////////////

int 
mtri_automata_initialize (void) {

  automata_list_head_ptr = NULL;

  return 0;
}




///////////////////////////////////////////////
//  mtri_create_automaton
///////////////////////////////////////////////

int
mtri_create_automaton (mtri_automaton_class_descriptor_t *class,
		       int                                thread,
		       int                                object)
{

  mtri_automaton_instance_descriptor_t *automaton_ptr;

#ifndef MODULE
  automaton_ptr = ( mtri_automaton_instance_descriptor_t *) 
    malloc (sizeof (mtri_automaton_instance_descriptor_t));
#else
  automaton_ptr = ( mtri_automaton_instance_descriptor_t *) 
    kmalloc (sizeof (mtri_automaton_instance_descriptor_t),
	     GFP_KERNEL);
  rtl_printf("(mtri_create_automaton) automaton_ptr = %x\n",
	     automaton_ptr);
#endif

  if (automaton_ptr == NULL) {
    return -1;
  }

  automaton_ptr->measuring     = FALSE;
  automaton_ptr->class         = class;
  automaton_ptr->thread        = thread;
  automaton_ptr->object        = object;
  automaton_ptr->current_state = 0;
  automaton_ptr->next          = automata_list_head_ptr;

  automata_list_head_ptr = automaton_ptr;

  return 0;
}





///////////////////////////////////////////////
//  mtri_match
///////////////////////////////////////////////

static int 
mtri_match (mtri_automaton_instance_descriptor_t *aut,
	    mtri_transition_descriptor_t         *tran_ptr,
	    mtri_input_tuple_t                   *ituple,
	    int                                  *matches,
	    int                                  *new_state)
{

  *matches = FALSE;

  switch (tran_ptr->kind) {

  case LAMBDA: {

    break;
  }

  case NORMAL: {

    if (tran_ptr->event_id != M_ANY_EVENT) {
      if (tran_ptr->event_id != ituple->event_id) {
	return 0;
      }
    }

    switch (tran_ptr->thread) {

    case ANY_THREAD: {
      break;
    }
    case SELF_THREAD: {
      if (aut->thread != ituple->thread) {
	return 0;
      }
      break;
    }
    case OTHER_THREAD: {
      if (aut->thread == ituple->thread) {
	return 0;
      }
      break;
    }
    } // end switch


    switch (tran_ptr->object) {

    case ANY_OBJECT: {
      break;
    }
    case SELF_OBJECT: {
      if (aut->object != ituple->object) {
	return 0;
      }
      break;
    }
    case OTHER_THREAD: {
      if (aut->object == ituple->object) {
	return 0;
      }
      break;
    }
    } // end switch


  default: {
    break;
  }

  }
  } // end switch

  *matches = TRUE;

  *new_state = tran_ptr->to_state;
  return 0;
}




///////////////////////////////////////////////
//  mtri_process_one_event
///////////////////////////////////////////////

int
mtri_process_one_event (mtri_automaton_instance_descriptor_t *aut,
			mtri_input_tuple_t                   *tuple,
			int                                  *measurement_done)
{
  int num_states;
  int matches;
  int new_state;
  int match_found = FALSE;

  mtri_state_kind_t new_state_kind;
  mtri_automaton_class_descriptor_t *class;
  mtri_transition_descriptor_t *tr;

  struct timespec segment_duration;

  *measurement_done = FALSE ;

  class = aut->class;
  num_states = class->num_states; // QUITAR?


  do {
    tr = class->transition [aut->current_state];

    while (tr != NULL) { 

      mtri_match (aut,
		  tr,
		  tuple,
		  &matches,
		  &new_state);
      
      if (matches == TRUE) {
	new_state_kind = class->state [new_state].kind;
	
	switch (new_state_kind) {
	  
	case IN: {
	  
	  if (aut->measuring == FALSE) {
	    
	    aut->measuring     = TRUE;
	    aut->in_segment    = TRUE;
	    aut->begin         = tuple->timestamp;
	    aut->end.tv_sec    = 0;
	    aut->end.tv_nsec   = 0;
	    aut->sum.tv_sec    = 0;
	    aut->sum.tv_nsec   = 0;
	    aut->partial_first = tuple->timestamp;
	    aut->events_count  = 0;
	    
	  } else {
	    if (aut->in_segment == FALSE) {
	      aut->in_segment = TRUE;
	      aut->partial_first = tuple->timestamp;
	    } else {
	      aut->events_count++;
	    }
	  }
	  break;
	} // end IN
	
	case OUT: {
	  if (aut->measuring == TRUE) {
	    if (aut->in_segment == TRUE) {
	      aut->in_segment = FALSE;

	      segment_duration = tuple->timestamp;
	      timespec_sub( &segment_duration , &(aut->partial_first) ); 
	      timespec_add(&(aut->sum), &segment_duration);

	      aut->events_count++;
	    }
	  }
	  break;
	} // end OUT

	case CANCEL: {
	  aut->measuring  = FALSE;
	  break;
	}
	// end CANCEL

	case END: {
	  if (aut->measuring == TRUE) {
	    
	    if (aut->in_segment == TRUE) {
	      segment_duration = tuple->timestamp;
	      timespec_sub( &segment_duration , &(aut->partial_first) ); 
	      timespec_add(&(aut->sum), &segment_duration);
	      aut->events_count++;
	    }

	    aut->end = tuple->timestamp;
	    aut->measuring  = FALSE;
	    aut->in_segment = FALSE;
	    *measurement_done = TRUE;
	  }
	} // end END
 
	default: {
	  break;
	}

	} // end switch (new_state_kind)
	
	aut->current_state = new_state;

	match_found = TRUE;

	break; 

      }
      
      tr = tr->next;
      
    } //end while
    


  } while (class->transition [aut->current_state]->kind == LAMBDA);
  
  if (match_found == FALSE && aut->in_segment == TRUE) {
    aut->events_count++;
  }


  return 0;
}




			    






///////////////////////////////////////////////
///////////////////////////////////////////////
//  INTERNALS
///////////////////////////////////////////////
///////////////////////////////////////////////

#define THREAD_MAX 128
#define MTRI_PENDING_ITUPLES_MAX THREAD_MAX
#define MTRI_METHOD_GET   0
#define MTRI_METHOD_TRY   1
#define MTRI_METHOD_TIMED 2


typedef enum {
  NO_MAP, SINGLE, MULTIPLE

} mtri_map_kind;


typedef struct {
  mtri_map_kind   kind;
  mtri_event_id_t m_event_id [MTRI_INDEX_MAX];

} mtri_map_element;


typedef struct {
  int        count;
  int        thread [THREAD_MAX];

} thread_bag_t;


typedef struct {
  int        count;
  struct {
    int type;
    int thread_id;
    int object_id;
  } item [THREAD_MAX];

} object_bag_t;


struct pending_node {
    mtri_input_tuple_t ituple;
    void * next;
};

struct {
  struct pending_node * first;
  struct pending_node * last;
} mtri_pending; 


static int mtri_initialized = FALSE;

static int MTRI_NUM_METRICS = NUM_SYSTEM_METRICS;

static thread_bag_t mtri_thread_bag;

static object_bag_t mtri_object_bag;

static mtri_map_element mtri_map [MTRI_TRACE_EVENTS_MAX];

static metrics_metricset_t  mtri_current_metrics_set;

static trace_id_t   mtri_current_trid; 

static int mtri_current_id;


///////////////////////////////////////////////
// mtri_timespec_less 
///////////////////////////////////////////////

static int
mtri_timespec_less (struct timespec *v1,
		    struct timespec *v2) 
{

  if (v1->tv_sec < v2->tv_sec) {
    return TRUE;
  } else if (v1->tv_sec > v2->tv_sec) {
    return FALSE;
  } else {
    return (v1->tv_nsec < v2->tv_nsec);
  }
}

///////////////////////////////////////////////
// mtri_internal_error
///////////////////////////////////////////////

static int
mtri_internal_error (char *s)
{
#ifndef MODULE
  fprintf (stderr, "INTERNAL ERROR: %s\n", s);
  exit (1);
#else 
  rtl_printf ("INTERNAL ERROR: %s\n", s);
  return -1;
#endif
}






///////////////////////////////////////////////
//  mtri_pending_initialize
///////////////////////////////////////////////

static int
mtri_pending_initialize (void)
{
  mtri_pending.first = NULL;
  mtri_pending.last  = NULL;

  return 0;
} 


///////////////////////////////////////////////
//  mtri_maps_initialize
///////////////////////////////////////////////

static int
mtri_maps_initialize (void)
{

  int i;

  for (i = 0; i < MTRI_TRACE_EVENTS_MAX; i++) {
    mtri_map [i].kind = NO_MAP;
  }

#include <metrics/mtri_initialize_maps_body.c>

  return 0;
} 




///////////////////////////////////////////////
//  mtri_internals_initialize
///////////////////////////////////////////////

static int
mtri_internals_initialize (trace_id_t                 trid,
			   const metrics_metricset_t *set)
{
  int i;
  int m_id;
  int ismember;
  int error = 0;

  mtri_automaton_class_descriptor_t *aclass_ptr;


  mtri_current_id = 0;

  // Automata module initialization

  mtri_automata_initialize ();


  // Maps initialization

  mtri_maps_initialize ();

  // Pending events

  mtri_pending_initialize ();


  // Register metrics set and trace

  for (i = 0; i < METRICS_MAX; i++) {
    mtri_current_metrics_set.s[i] = set->s[i];
  }

  mtri_current_trid = trid;


  // SYSTEM_WIDE automata creation

  for (m_id = 0; m_id < NUM_SYSTEM_METRICS; m_id++) {

    metrics_metricset_ismember (m_id, set, &ismember);
    
    if (ismember == TRUE) {
      
      aclass_ptr = &(mtri_metric [m_id].automaton);
      
      if (aclass_ptr->kind == SYSTEM_WIDE) {
	error = mtri_create_automaton (aclass_ptr, MTRI_NULL_THREAD, MTRI_NULL_OBJECT);
	if (error) {
#ifdef MODULE
	  rtl_printf("(mtri_internals_initialize) Error creating automaton\n");
#endif
	  return -1;
	}
      }
    }
    
  } // end for

  // Thread bag initialization
  mtri_thread_bag.count = 0;



  return 0;
} 



///////////////////////////////////////////////
//  mtri_register_thread
///////////////////////////////////////////////


static int 
mtri_register_thread (int thread_id) {
  
  int i;
  int count;
  int m_id;
  int ismember;

  mtri_automaton_class_descriptor_t *aclass_ptr;


  count = mtri_thread_bag.count;
  
  for (i = 0; i < count; i++) {
    if (mtri_thread_bag.thread [i] == thread_id) {
      break;
    }
  }
  
  if (i == count) {

    // New thread

    mtri_thread_bag.thread [i] = thread_id;
    mtri_thread_bag.count++;

    for (m_id = 0; m_id < MTRI_NUM_METRICS; m_id++) {
      
      metrics_metricset_ismember (m_id, &mtri_current_metrics_set, &ismember);
      
      if (ismember == TRUE) {
	
	aclass_ptr = &(mtri_metric [m_id].automaton);
	
	if (aclass_ptr->kind == THREAD_RELATED) {
	  mtri_create_automaton (aclass_ptr, thread_id, MTRI_NULL_OBJECT);
	}
      }
      
    } // end for
  }
  
  return 0;
}


///////////////////////////////////////////////
//  mtri_register_object
///////////////////////////////////////////////


static int 
mtri_register_object (int thread_id,
		      int type,
		      int object_id) {
  
  int i;
  int count;
  int m_id;
  int ismember;

  mtri_automaton_class_descriptor_t *aclass_ptr;



  count = mtri_object_bag.count;
  
  for (i = 0; i < count; i++) {
    if (mtri_object_bag.item [i].thread_id == thread_id && 
	mtri_object_bag.item [i].type      == type && 
	mtri_object_bag.item [i].object_id == object_id) {
      break;
    }
  }
  


  if (i == count) {

    // New Object

    mtri_object_bag.item [i].type = type;
    mtri_object_bag.item [i].object_id = object_id;
    mtri_object_bag.item [i].thread_id = thread_id;
    mtri_object_bag.count++;


    for (m_id = 0; m_id < MTRI_NUM_METRICS; m_id++) {
      
      metrics_metricset_ismember (m_id, &mtri_current_metrics_set, &ismember);
      
      if (ismember == TRUE) {
	
	aclass_ptr = &(mtri_metric [m_id].automaton);
	
	if (aclass_ptr->kind == type) {
	  mtri_create_automaton (aclass_ptr, thread_id, object_id);
	}
      }
      
    } // end for
  }
  

  return 0;
}


/* static int */
/* show_pendings (void) { */

/*   struct pending_node * ptr; */
  
/*   return 0; */

/*   fprintf (stderr, "PENDINGS: "); */
/*   ptr  = mtri_pending.first; */
/*   while (ptr != NULL) { */
/*     fprintf (stderr, "(%d,%d)", ptr->ituple.event_id, ptr->ituple.thread); */
/*     ptr =  ptr->next; */
/*   } */
/*   fprintf (stderr, "\n"); */

/*   return 0; */
/* } */


///////////////////////////////////////////////
//  mtri_add_pending_ituple
///////////////////////////////////////////////

static int
mtri_add_pending_ituple (mtri_input_tuple_t     *ituple) {

  struct pending_node * new_node;
  struct pending_node * ptr;
  struct pending_node * prev;
  
#ifndef MODULE
  new_node = (struct pending_node *) malloc (sizeof (struct pending_node));
#else
  new_node = (struct pending_node *) kmalloc (sizeof (struct pending_node),
					      GFP_KERNEL);
#endif
  new_node->ituple = *ituple;
  
  ptr  = NULL;
  prev = NULL;

  if (mtri_pending.first != NULL) {
    ptr = mtri_pending.first;
    while (mtri_timespec_less (&(ptr->ituple.timestamp), &(ituple->timestamp))) {
      prev = ptr;
      ptr =  ptr->next;
      if (ptr == NULL) {
	break;
      } 
    }
  }

  if (mtri_pending.first == NULL) {
    
    mtri_pending.first = new_node;
    mtri_pending.last  = new_node;
    new_node->next = NULL;

  } else if (prev == NULL) {

    new_node->next = mtri_pending.first;
    mtri_pending.first = new_node;

  } else {
    new_node->next = prev->next;
    prev->next = new_node;
    if (ptr == NULL) {
      mtri_pending.last = new_node;
    }
  }

  return 0;
}


///////////////////////////////////////////////
//  mtri_first_pending_ituple
///////////////////////////////////////////////

static int
mtri_first_pending_ituple (mtri_input_tuple_t     * ituple,
			   int                    * available) {

  if (mtri_pending.first == NULL) {
    *available = FALSE;
    return 0;
  }

  *available = TRUE;
  *ituple = (mtri_pending.first)->ituple;

  return 0;
}



///////////////////////////////////////////////
//  mtri_remove_first_pending_ituple
///////////////////////////////////////////////

static int
mtri_remove_first_pending_ituple (void) {

  struct pending_node * ptr;

  if (mtri_pending.first == NULL) {
   return 0;
  }

  ptr = (mtri_pending.first)->next;

#ifndef MODULE
  free (mtri_pending.first);
#else
  kfree (mtri_pending.first);
#endif

  mtri_pending.first = ptr;

  return 0;

}


///////////////////////////////////////////////
//  mtri_process_release_events
///////////////////////////////////////////////

static int
mtri_process_release_events (posix_trace_event_info *event,
			     void                   *data)
{
  posix_trace_sysevent_data * specific_data;
  struct timespec             until;
  mtri_input_tuple_t          ituple;

  specific_data = (posix_trace_sysevent_data *) data;

  switch (event->posix_event_id) {
  case POSIX_TRACE_SYSCALL: {
    switch (specific_data->index) {
    case SYSCALL_CLOCK_NANOSLEEP: {
      until = specific_data->other_data.delay_until;
      ituple.event_id  = M_CLOCK_NANOSLEEP_RELEASE;
      ituple.timestamp = specific_data->other_data.delay_until;
      ituple.thread    = specific_data->task;
      ituple.object    = MTRI_NULL_OBJECT;
      mtri_add_pending_ituple (&ituple);
      break;
    }
    }
  }
  }

  specific_data = (posix_trace_sysevent_data *) data;

  return 0;
}



///////////////////////////////////////////////
//  mtri_trace_event_to_input_tuple
///////////////////////////////////////////////

static int 
mtri_trace_event_to_input_tuple (posix_trace_event_info *event,
				 void                   *data,
				 mtri_input_tuple_t     *ituple)
{

  mtri_map_kind map_kind;
  posix_trace_sysevent_data *specific_data;



  ituple->timestamp = event->posix_timestamp;

  specific_data = (posix_trace_sysevent_data *) data;

  map_kind = mtri_map [event->posix_event_id].kind;

  switch (map_kind) {
  case NO_MAP: {
    ituple->event_id = M_NO_EVENT;
    ituple->thread   = MTRI_NULL_THREAD;
    return 0;
  }

  case SINGLE: {
    ituple->event_id = mtri_map [event->posix_event_id].m_event_id [0];
    break;
  }
  case MULTIPLE: {
    ituple->event_id = mtri_map [event->posix_event_id].m_event_id [specific_data->index];
    break;
  }
  }


  ituple->thread = MTRI_NULL_THREAD;
  ituple->object = MTRI_NULL_OBJECT;

  if (mtri_register [ituple->event_id].thread == THREAD_RELATED) {
    ituple->thread = specific_data->task;  
    mtri_register_thread (ituple->thread);
  } 
    
  if (mtri_register [ituple->event_id].object != NO_KIND) {
    ituple->thread = specific_data->task;    
    switch (mtri_register [ituple->event_id].object) {
    case MUTEX_RELATED: {
      ituple->object = specific_data->other_data.mutex_data.mutex;
      break;
    }
    case COND_RELATED: {
      ituple->object = specific_data->other_data.cond_data.cond;
      break;
    }
    }
    mtri_register_object (ituple->thread, mtri_register [ituple->event_id].object, ituple->object);
  } 


  mtri_process_release_events (event, data);

  if (ituple->event_id == M_INVALID_EVENT) {
    mtri_internal_error ("M_INVALID_EVENT");
  }

  return 0;
}







///////////////////////////////////////////////
//  mtri_getnext_measurement
///////////////////////////////////////////////

// globals used only by this function

int                events_processed = 0;
int                ituple_in_curse = FALSE;
int                have_saved_ituple = FALSE;
mtri_input_tuple_t current_ituple;
mtri_input_tuple_t saved_ituple;
mtri_automaton_instance_descriptor_t *current_automaton_ptr;

static int
mtri_getnext_measurement (int                    method,
			  metrics_measurement_t *result,
			  int                   *unavailable,
			  const struct timespec       *abs_timeout)
{

  int                           res = 0;
  int                           no_more_events;
  int                           measurement_done;
  struct posix_trace_event_info event;
  char                          data[256];
  size_t                        data_len;
  mtri_input_tuple_t            last_tuple;
  mtri_input_tuple_t            temp_ituple;
  int                           processing_last_tuple = FALSE;
  int                           available;


  *unavailable = FALSE;



  do {
    if (ituple_in_curse == FALSE) {

      if (have_saved_ituple == TRUE) {
	current_ituple = saved_ituple;
      } else {
	switch (method) {
	case MTRI_METHOD_GET: {
	  res = posix_trace_getnext_event(mtri_current_trid, &event, 
					  data, sizeof(data), 
					  &data_len, &no_more_events);
	  break;
	}
#ifdef MODULE
	case MTRI_METHOD_TRY: {
	  res = posix_trace_trygetnext_event(mtri_current_trid, &event, 
					     data, sizeof(data), 
					     &data_len, &no_more_events);
	  break;
	}      
	case MTRI_METHOD_TIMED: {
	  res = posix_trace_timedgetnext_event(mtri_current_trid, &event, 
					       data, sizeof(data), 
					       &data_len, &no_more_events,
					       abs_timeout);
	  break;
	} 
#endif     
	} // end switch
	
	
	if (res != 0) {
	  return res;
	}
	
	if (no_more_events == TRUE) {
	  processing_last_tuple = TRUE;
	  last_tuple.event_id = M_LAST_EVENT;
	  last_tuple.timestamp = current_ituple.timestamp;
	  
	  current_ituple = last_tuple;
	  
	} else {  
	  mtri_trace_event_to_input_tuple (&event, &data, &current_ituple);
	}
      } 
	

      have_saved_ituple = FALSE;
      mtri_first_pending_ituple (&temp_ituple, &available);
      if (available == TRUE) {
	if (mtri_timespec_less (&temp_ituple.timestamp, &current_ituple.timestamp) == TRUE) {
	  saved_ituple = current_ituple;
	  current_ituple = temp_ituple;
	  mtri_remove_first_pending_ituple ();
	  have_saved_ituple = TRUE;
	}
      }
           
      current_automaton_ptr = automata_list_head_ptr;
      ituple_in_curse = TRUE;
    }


    while (current_automaton_ptr != NULL) {


      mtri_process_one_event (current_automaton_ptr,
			      &current_ituple,
			      &measurement_done);
      
      if (measurement_done == TRUE) {
	result->metric_id = current_automaton_ptr->class->metric_id;

	result->thread_id = MTRI_NULL_THREAD;
	result->object_id = MTRI_NULL_OBJECT;
	if (current_automaton_ptr->class->kind == SYSTEM_WIDE) {
	  // nothing to do
	} else if (current_automaton_ptr->class->kind == THREAD_RELATED) {
	  result->thread_id = current_automaton_ptr->thread;
	} else {
	  result->thread_id = current_automaton_ptr->thread;
	  result->object_id = current_automaton_ptr->object;	  
	}


	result->duration     = current_automaton_ptr->sum;
	result->begin        = current_automaton_ptr->begin;
	result->end          = current_automaton_ptr->end;
	result->events_count = current_automaton_ptr->events_count;
	result->id           = mtri_current_id;

	mtri_current_id++;

	// A measurement has been found.


	current_automaton_ptr = current_automaton_ptr->next;
	return 0;
      }

      current_automaton_ptr = current_automaton_ptr->next;

    } //end while

    // Current tuple has been completely processed and 
    // no measurement has been found.
    
    ituple_in_curse = FALSE;

    if (processing_last_tuple == TRUE) {
      *unavailable = TRUE;
      return 0;
    }
    
  } while (TRUE);
}


///////////////////////////////////////////////
///////////////////////////////////////////////
//  INTERFACES
///////////////////////////////////////////////
///////////////////////////////////////////////



///////////////////////////////////////////////
//  metrics_metric_open
///////////////////////////////////////////////

int
metrics_metric_open (const char          *metric_name,
		     metrics_metric_id_t *metric_id) 
{

  int i;

  if (metric_name == NULL) {
    return EINVAL;
  }
  
  *metric_id = MTRI_NO_METRIC;
  for (i = 0; i < NUM_SYSTEM_METRICS; i++) {
    if (strcmp (metric_name, mtri_metric [i].name) == 0) {
      *metric_id = i;
      break;
    }
  }

  if (*metric_id != MTRI_NO_METRIC) {
    return 0;
  } else {
    return EINVAL;
  }
}




///////////////////////////////////////////////
//  metrics_metric_getname
///////////////////////////////////////////////

int
metrics_metric_getname (metrics_metric_id_t metric_id,
			char               *metric_name)
{
  if (metric_id < 0 || metric_id >= MTRI_NUM_METRICS) {
    return EINVAL;
  }

  if (metric_name == NULL) {
    return EINVAL;
  }
    
  strcpy (metric_name, mtri_metric [metric_id].name);

  return 0;
}



///////////////////////////////////////////////
//  metrics_metric_getdescription
///////////////////////////////////////////////

int
metrics_metric_getdescription (metrics_metric_id_t  metric_id,
			       char                *metric_description)
{

  if (metric_id < 0 || metric_id >= MTRI_NUM_METRICS) {
    return EINVAL;
  }

  if (metric_description == NULL) {
    return EINVAL;
  }
    
  strcpy (metric_description, mtri_metric [metric_id].description);

  return 0;
}




///////////////////////////////////////////////
//  metrics_metricset_empty
///////////////////////////////////////////////

int
metrics_metricset_empty (metrics_metricset_t *set)
{
  int i;

  if (set == NULL) {
    return EINVAL;
  }

  for (i = 0; i < METRICS_MAX; i++) {
    set->s[i] = FALSE;
  }
  set->current_pos = 0;

  return 0;
}




///////////////////////////////////////////////
//  metrics_metricset_fill
///////////////////////////////////////////////

int
metrics_metricset_fill (metrics_metricset_t *set)
{
  int i;

  if (set == NULL) {
    return EINVAL;
  }

  for (i = 0; i < METRICS_MAX; i++) {
    set->s[i] = TRUE;
  }
  set->current_pos = 0;

  return 0;
}



///////////////////////////////////////////////
//  metrics_metricset_add
///////////////////////////////////////////////

int
metrics_metricset_add (metrics_metric_id_t  metric_id,
		       metrics_metricset_t *set)
{
  if (metric_id < 0 || metric_id >= MTRI_NUM_METRICS) {
    return EINVAL;
  }

  if (set == NULL) {
    return EINVAL;
  }
    
  set->s[metric_id] = TRUE;

  return 0;
}


///////////////////////////////////////////////
//  metrics_metricset_del
///////////////////////////////////////////////

int
metrics_metricset_del (metrics_metric_id_t  metric_id,
		       metrics_metricset_t *set)
{
  if (metric_id < 0 || metric_id >= MTRI_NUM_METRICS) {
    return EINVAL;
  }

  if (set == NULL) {
    return EINVAL;
  }
    
  set->s[metric_id] = FALSE;

  return 0;
}




///////////////////////////////////////////////
//  metrics_metricset_ismember
///////////////////////////////////////////////

int
metrics_metricset_ismember (metrics_metric_id_t        metric_id,
			    const metrics_metricset_t *set,
			    int                       *ismember)
{
  if (metric_id < 0 || metric_id >= MTRI_NUM_METRICS) {
    return EINVAL;
  }

  if (set == NULL) {
    return EINVAL;
  }
    
  *ismember = set->s[metric_id];

  return 0;
}



///////////////////////////////////////////////
//  metrics_metricset_iterator_getnext 
///////////////////////////////////////////////

int 
metrics_metricset_iterator_getnext (metrics_metricset_t *set,
				    metrics_metric_id_t       *metric_id,
				    int                       *unavailable)
{
  metrics_metric_id_t i;
  
  *unavailable = FALSE;

  for (i = set->current_pos; i < MTRI_NUM_METRICS; i++) {
    if (set->s[i] == TRUE) {
      *metric_id = i;
      set->current_pos = i + 1;
      return 0;
    }
  }
  *unavailable = TRUE;

  return 0;
}




///////////////////////////////////////////////
//  metrics_metricset_iterator_rewind 
///////////////////////////////////////////////

int
metrics_metricset_iterator_rewind (metrics_metricset_t *set)
{
  set->current_pos = 0;
  return 0;
}




///////////////////////////////////////////////
//  metrics_init
///////////////////////////////////////////////

int
metrics_init (trace_id_t                 trid,
	      const metrics_metricset_t *set)
{
  trace_attr_t attr;
  int          res;



  res = posix_trace_get_attr (trid, &attr);

  if (res != 0) {
    return EINVAL;
  }

  if (set == NULL) {
    return EINVAL;
  }

  if (mtri_initialized == TRUE) {
    return EPERM;
  } 


  mtri_internals_initialize (trid, set);
  
  mtri_initialized = TRUE;
    
  return 0;
}





///////////////////////////////////////////////
//  metrics_destroy
///////////////////////////////////////////////

int
metrics_destroy (void)
{

  // PENDING

  return 0;
}




///////////////////////////////////////////////
//  metrics_getnext_measurement
///////////////////////////////////////////////

int
metrics_getnext_measurement (metrics_measurement_t *result,
			     int                   *unavailable)
{

  int res;

  if (result == NULL) {
    return EINVAL;
  }

  if (unavailable == NULL) {
    return EINVAL;
  }

  if (mtri_initialized == FALSE) {
    return EPERM;
  }


  res = mtri_getnext_measurement (MTRI_METHOD_GET, result, unavailable, NULL);

  return res;
}


///////////////////////////////////////////////
//  metrics_trygetnext_measurement
///////////////////////////////////////////////

int
metrics_trygetnext_measurement (metrics_measurement_t *result,
				int                   *unavailable)
{

  int res;

  if (result == NULL) {
    return EINVAL;
  }

  if (unavailable == NULL) {
    return EINVAL;
  }

  if (mtri_initialized == FALSE) {
    return EPERM;
  }

  res = mtri_getnext_measurement (MTRI_METHOD_TRY, result, unavailable, NULL);

  return 0;
}




///////////////////////////////////////////////
//  metrics_trygetnext_measurement
///////////////////////////////////////////////

int
metrics_timedgetnext_measurement (metrics_measurement_t *result,
				  int                   *unavailable,
				  const struct timespec *abs_timeout)
{
  int res;


  if (result == NULL) {
    return EINVAL;
  }

  if (unavailable == NULL) {
    return EINVAL;
  }

  if (mtri_initialized == FALSE) {
    return EPERM;
  }

  res = mtri_getnext_measurement (MTRI_METHOD_TIMED, result, unavailable, abs_timeout);

  return 0;
}



#ifdef MODULE
MODULE_LICENSE("GPL");
int init_module(void) {
  return 0;
}

void cleanup_module(void) {
}
#endif

