#include <trace.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include <metrics.h>

#define NORMALIZE 1
#define PATH_LENGTH_MAX  512
#define MAX_METRICS      30


//////////////////////////////////////////////////////////////
//  usage
//////////////////////////////////////////////////////////////

static void
usage (char * s)
{
  fprintf (stderr, "usage: %s [-i input] [-o output] [-s metrics] [--help]\n", s);
  fprintf (stderr, "\t -i input   : input  file (default standard input)\n");
  fprintf (stderr, "\t -d output  : output file (default standard output)\n");
  fprintf (stderr, "\t -s metrics : list of metrics names to get \n");
  fprintf (stderr, "\t --help     : prints this message and exit\n");
  exit (1);
}


//////////////////////////////////////////////////////////////
//  minus
//////////////////////////////////////////////////////////////

int
minus (struct timespec  to,
       struct timespec  from,
       struct timespec *result) 

{
  if (to.tv_nsec >= from.tv_nsec) {
    result->tv_sec = to.tv_sec - from.tv_sec;
    result->tv_nsec = to.tv_nsec - from.tv_nsec;
  } else {
    result->tv_sec = to.tv_sec - from.tv_sec + 1;
    result->tv_nsec = to.tv_nsec + 1000000000 - from.tv_nsec;
  }

  return 0;
}


//////////////////////////////////////////////////////////////
//  main
//////////////////////////////////////////////////////////////

int
main (int argc, char *argv[]) {

  int res;

  metrics_metricset_t    set;
  metrics_metric_id_t    metric_id;
  metrics_measurement_t  result;
  int                    unavailable = FALSE;
  char                   name [METRICS_NAME_MAX];
  char                   description [METRICS_DESCRIPTION_MAX];

  trace_id_t                    trid;
  int                           no_more_events;
  struct posix_trace_event_info event;
  char                          data[256];
  size_t                        data_len;

  struct timespec base;
  struct timespec begin_normalized;
  struct timespec end_normalized;

  time_t today;
  char   hostname [128];


  char   inputfilename  [PATH_LENGTH_MAX] = "";
  char   outputfilename [PATH_LENGTH_MAX] = "";
  int    input;
  FILE * output;
  int    option_input  = FALSE;
  int    option_output = FALSE;
  int    option_select = FALSE;
  int    i;
  int    error;

  metrics_metric_id_t   metric_selected      [METRICS_MAX];
  char                 *metric_selected_name [METRICS_MAX];
  int                   selected_counter = 0;

  struct timespec               time_spec;
  char                          str[TRACE_NAME_MAX];
  char                          polstr[TRACE_NAME_MAX];
  int                           pol;
  trace_attr_t                  trace_attr;


  for (i = 1; i < argc; i++) {

    if (option_input == TRUE) {
      strcpy (inputfilename, argv [i]);
      option_input = FALSE;

    } else if (option_output == TRUE) {
      strcpy (outputfilename, argv [i]);
      option_output = FALSE;

    } else if (option_select == TRUE) {
      if ( argv [i][0] == '-' ) {
	i--;
	option_select = FALSE;
      } else {
	metric_selected_name [selected_counter] = argv [i];
	selected_counter++;
      }
    } else if (strcmp (argv [i], "-o") == 0) {
      option_output = TRUE;

    } else if (strcmp (argv [i], "-i") == 0) {
      option_input = TRUE;

    } else if (strcmp (argv [i], "-s") == 0) {
      option_select = TRUE;

    } else if (strcmp (argv [i], "--help") == 0) {
      usage (argv [0]);

    } else {
      usage (argv [0]);
    }      
  }       
    
  if (strcmp (inputfilename, "") == 0) {
    input = STDIN_FILENO;
    input = 0;

  } else {
    input = open (inputfilename, O_RDWR);
    
    if (input < 0) {
      fprintf (stderr, "error opening input file %s>>\n", inputfilename);
      perror ("que pasa");
      usage (argv [0]);
    }
  }

  if (strcmp (outputfilename, "") == 0) {
    output = stdout;

  } else {
    output = fopen (outputfilename, "w+");
    
    if (output == NULL) {
      fprintf (stderr, "error opening output file %s\n", outputfilename);
      usage (argv [0]);
    }
  }

  for (i = 0; i < selected_counter; i++) {
    res = metrics_metric_open (metric_selected_name [i], 
			       &metric_selected [i]);
    if (res != 0) {
      fprintf (stderr, "metric %s does not exist\n", metric_selected_name [i]);
      usage (argv [0]);
    }      
  }
  
  time (&today);

  gethostname (hostname, 128);


  posix_trace_open (input, &trid);


  res = posix_trace_getnext_event(trid, &event, 
				  data, sizeof(data), 
				  &data_len, &no_more_events);
  
  base = event.posix_timestamp;

  if (selected_counter == 0) {
    metrics_metricset_fill (&set);

  } else {

    metrics_metricset_empty (&set); 
    for (i = 0; i < selected_counter; i++) {
      metrics_metricset_add (metric_selected [i], &set);
    }      
  }    

  fprintf (output, "<METRICSREPORT>\n");

  error = posix_trace_get_attr(trid, &trace_attr);



  fprintf (output, "<HEADER>\n");
  
  fprintf (output, "<CREATED> %s </CREATED>\n", ctime (&today));
  fprintf (output, "<HOST> %s </HOST>\n", hostname);
  fprintf (output, "<USER> %s </USER>\n", getlogin ());
  fprintf (output, "<METRICSREQUESTED>\n");
  metrics_metricset_iterator_getnext (&set, &metric_id, &unavailable);
  while (unavailable == FALSE) {
    res = metrics_metric_getname (metric_id, name);
    res = metrics_metric_getdescription (metric_id, description);
    if (res == 0) {
      fprintf (output, "  <METRIC><METRICNAME>%s</METRICNAME><METRICDESCRIPTION>%s</METRICDESCRIPTION><METRICID>%d</METRICID></METRIC>\n", name, description, metric_id);
    }
    metrics_metricset_iterator_getnext (&set, &metric_id, &unavailable);
  }
  fprintf (output, "</METRICSREQUESTED>\n");
  fprintf (output, "</HEADER>\n\n");

  fprintf (output, "<TRACE-ATTRIBUTES>\n");
  error = posix_trace_attr_getgenversion(&trace_attr, str);
  fprintf(output, "      <GENERATION-VERSION>%s</GENERATION-VERSION>\n", 
	  str);
  
  error = posix_trace_attr_getname(&trace_attr, str);
  fprintf(output, "      <TRACE-NAME>%s</TRACE-NAME>\n", str);

  error = posix_trace_attr_getcreatetime(&trace_attr, &time_spec);
  fprintf(output, "      <TRACE-CREATION-TIME>%ld.%09ld</TRACE-CREATION-TIME>\n", 
	  time_spec.tv_sec, time_spec.tv_nsec);

  error = posix_trace_attr_getclockres(&trace_attr, &time_spec);
  fprintf(output, "      <CLOCK-RESOLUTION>%ld.%09ld</CLOCK-RESOLUTION>\n", 
	  time_spec.tv_sec, time_spec.tv_nsec);

  error = posix_trace_attr_getstreamfullpolicy(&trace_attr, &pol);
  switch (pol) {
  case POSIX_TRACE_LOOP:
    sprintf(polstr, "POSIX_TRACE_LOOP");
    break;
  case POSIX_TRACE_UNTIL_FULL:
    sprintf(polstr, "POSIX_TRACE_UNTIL_FULL");
    break;
  case POSIX_TRACE_FLUSH:
    sprintf(polstr, "POSIX_TRACE_FLUSH");
    break;
  default:
    sprintf(polstr, "#UNKNOWN#");
  }
  fprintf(output, "      <STREAM-FULL-POLICY>%s</STREAM-FULL-POLICY>\n", 
	  polstr);

  fprintf (output, "</TRACE-ATTRIBUTES>\n");



  fprintf (output, "<MEASUREMENTS>\n");




  metrics_init (trid, &set);

  metrics_getnext_measurement (&result, &unavailable);

  while (unavailable == FALSE) {
    
    begin_normalized = result.begin;
    end_normalized = result.end;
    
    if (NORMALIZE == TRUE) {
      timespec_sub (&end_normalized, &base);
      timespec_sub (&begin_normalized, &base);
    }
    
    metrics_metric_getname (result.metric_id, name);
    
    fprintf (output, "<MEASUREMENT>\n");
    fprintf (output, "  <METRICID>%d</METRICID>\n",     result.metric_id);
    fprintf (output, "  <METRICNAME>%s</METRICNAME>\n", name);
    fprintf (output, "  <THREAD>%d</THREAD>\n",           result.thread_id);
    fprintf (output, "  <OBJECT>%d</OBJECT>\n",           result.object_id);
    fprintf (output, "  <DURATION>%ld.%09ld</DURATION>\n",  result.duration.tv_sec, result.duration.tv_nsec);
    fprintf (output, "  <BEGIN>%ld.%09ld</BEGIN>\n",        begin_normalized.tv_sec, begin_normalized.tv_nsec);
    fprintf (output, "  <END>%ld.%09ld</END>\n",            end_normalized.tv_sec, end_normalized.tv_nsec);
    fprintf (output, "  <EVCOUNT>%d</EVCOUNT>\n",       result.events_count);
    fprintf (output, "  <ID>%d</ID>\n",                result.id);
    fprintf (output, "</MEASUREMENT>\n\n");

    metrics_getnext_measurement (&result, &unavailable); 
  }

  fprintf (output, "</MEASUREMENTS>\n");

  fprintf (output, "</METRICSREPORT>\n");

  return 0;
}
