/*
   FILE:
						       	
	test_ptrace_api3_module.c
								       	
   DESCRIPTION:							       	

        Test 3 for POSIX Tracing API in RT-Linux.
	
	This test checks the online trace and retrieval of events.

	In particular, the init_module traces some events and 
	creates three tasks which check one different retrieval
	function each:
 
	  1) Task "reader_try" tests the trygetnext_event function.
	     It executes this retrieval function twice with only one
	     event in the stream: it should retrieve the event and
	     then fail to retrieve the second time.

	  2) Task "reader_timed" tests the timedgetnext_event
	     function.  It executes this retrieval function twice with
	     only one event in the stream: it should retrieve the
	     event and then fail (timeout) to retrieve the second
	     time.

	  3) Task "reader_get" tests the getnext_event function.  It
	     executes this retrieval function twice with only one
	     event in the stream: it should retrieve the event and
	     then fail (suspend forever) to retrieve the second time.

				       				
   Copyright (C) 2004
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        September 2004

*/

#include <rtl.h>
#include <time.h>
#include <trace.h>
#include <asm/errno.h>


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andres Terrasa");
MODULE_DESCRIPTION("API Test 3 for POSIX Tracing");


#define TEST_NAME "ptrace_test_api3"


#define MY_EVENT_DATA_SIZE   (size_t)    64
#define MY_STREAM_SIZE       (size_t)  4096

#define MY_STRING            "Please do not feed the programmers"


static int            test_failed = 0;
static int            timeout     = 0;
static int            woke_first  = 0;
static int            init_module_finished = 0;
static pthread_t      reader1, reader2, reader3;
static trace_id_t     trid;




void *reader_try(void *dummy) {
  int             error;
  struct posix_trace_event_info event;
  char            event_name[TRACE_NAME_MAX];
  char            data[MY_EVENT_DATA_SIZE];
  size_t          datalen;
  int             unavailable;

  /* Retrieve the first event and check that it's correct */
  error = posix_trace_trygetnext_event(trid, &event, &data, sizeof(data), 
				       &datalen, &unavailable);
  if (error) {
    rtl_printf("%s: [BROK] :posix_trace_trygetnext_event returned %d\n", 
	       TEST_NAME, error);
    test_failed = 1;
    return (void *) -1;
  }

  if (unavailable) {
    rtl_printf("%s: [BROK] :Stream empty when it should be an event\n", 
	       TEST_NAME);
    test_failed = 1;
    return (void *) -1;
  } 

  posix_trace_eventid_get_name (trid, event.posix_event_id, event_name);
  if (strcmp(event_name,"user event string")) {
    rtl_printf("%s: [BROK] :Retrieved event of wrong type\n", 
	       TEST_NAME);
    test_failed = 1;
    return (void *) -1;
  }
  
  if (strcmp((char *) data, MY_STRING)) {
      rtl_printf("%s: [BROK] :Retrieved event whith wrong data\n", 
	       TEST_NAME);
    test_failed = 1;
    return (void *) -1;
  }


  /* Retrieve the second event (it should not be any) */
  error = posix_trace_trygetnext_event(trid, &event, &data, sizeof(data), 
				       &datalen, &unavailable);

  if (error) {
    rtl_printf("%s: [BROK] :posix_trace_trygetnext_event returned %d\n", 
	       TEST_NAME, error);
    test_failed = 1;
    return (void *) -1;
  }

  if (! unavailable) {
    rtl_printf("%s: [BROK] :There was an extra event in the stream\n", 
	       TEST_NAME);
    test_failed = 1;
    return (void *) -1;
  }

  return (void *) 0;
}






void *reader_timed(void *dummy) {
  int             error;
  struct posix_trace_event_info event;
  char            data[64];
  size_t          datalen;
  char            event_name[TRACE_NAME_MAX];
  int             unavailable;
  struct timespec abs_timeout;

  /* Set the absolute timout to now + 500 usec */
  clock_gettime(CLOCK_MONOTONIC, &abs_timeout);
  timespec_add_ns(&abs_timeout, 500000);

  /* Retrieve the first event and check that it's correct */
  error = posix_trace_timedgetnext_event(trid, &event, &data, sizeof(data), 
					 &datalen, &unavailable, &abs_timeout);
  if (error) {
    rtl_printf("%s: [BROK] :posix_trace_timedgetnext_event returned %d\n", 
	       TEST_NAME, error);
    test_failed = 1;
    return (void *) -1;
  }

  if (unavailable) {
    rtl_printf("%s: [BROK] :Stream empty when it should be an event\n", 
	       TEST_NAME);
    test_failed = 1;
    return (void *) -1;
  } 

  posix_trace_eventid_get_name (trid, event.posix_event_id, event_name);
  if (strcmp(event_name,"user event string")) {
    rtl_printf("%s: [BROK] :Retrieved event of wrong type\n", 
	       TEST_NAME);
    test_failed = 1;
    return (void *) -1;
  }
  
  if (strcmp((char *) data, MY_STRING)) {
      rtl_printf("%s: [BROK] :Retrieved event whith wrong data\n", 
	       TEST_NAME);
    test_failed = 1;
    return (void *) -1;
  }

  /* Set the absolute timout to now + 500 usec */
  clock_gettime(CLOCK_MONOTONIC, &abs_timeout);
  timespec_add_ns(&abs_timeout, 500000);

  /* Retrieve the second event (now it SHOULD timeout) */
  error = posix_trace_timedgetnext_event(trid, &event, &data, sizeof(data), 
					 &datalen, &unavailable, &abs_timeout);

  if (error == -ETIMEDOUT) {
    timeout = 1;
  } else {
    rtl_printf("%s: [BROK] :NO timeout when retrieving an event\n", 
	       TEST_NAME, error);
    test_failed = 1;
    return (void *) -1;
  }

  if (! unavailable) {
    rtl_printf("%s: [BROK] :There was an extra event in the stream\n", 
	       TEST_NAME);
    test_failed = 1;
    return (void *) -1;
  }

  return (void *) 0;
}






void *reader_get(void *dummy) {
  int             error;
  struct posix_trace_event_info event;
  char            data[64];
  char            event_name[TRACE_NAME_MAX];
  size_t          datalen;
  int             unavailable;


  /* Retrieve the first event and check that it's correct */
  /* Set the test_fail just in case we do not wake up */
  test_failed = 1;
  error = posix_trace_getnext_event(trid, &event, &data, sizeof(data), 
				    &datalen, &unavailable);

  test_failed = 0; // We woke up after all...
  woke_first  = 1;

  if (error) {
    rtl_printf("%s: [BROK] :posix_trace_getnext_event returned %d\n", 
	       TEST_NAME, error);
    test_failed = 1;
    return (void *) -1;
  }

  if (unavailable) {
    rtl_printf("%s: [BROK] :Stream empty when it should be an event\n", 
	       TEST_NAME);
    test_failed = 1;
    return (void *) -1;
  } 

  posix_trace_eventid_get_name (trid, event.posix_event_id, event_name);
  if (strcmp(event_name,"user event string")) {
    rtl_printf("%s: [BROK] :Retrieved event of wrong type\n", 
	       TEST_NAME);
    test_failed = 1;
    return (void *) -1;
  }
  
  if (strcmp((char *) data, MY_STRING)) {
      rtl_printf("%s: [BROK] :Retrieved event whith wrong data\n", 
	       TEST_NAME);
    test_failed = 1;
    return (void *) -1;
  }

  /* Retrieve the second event (now it should NEVER wake up) */
  error = posix_trace_getnext_event(trid, &event, &data, sizeof(data), 
					 &datalen, &unavailable);

  if (! init_module_finished) {
    rtl_printf("%s: [BROK] :There was an extra event in the stream\n", 
	       TEST_NAME);
    test_failed = 1;
    return (void *) -1;
  }

  // Else we are awaking from clenaup module
  return (void *) 0;


}






int init_module(void) {

  trace_attr_t   attr;
  int            error;
  int            do_nothing;
  trace_event_id_t event_string_id;
  trace_event_set_t  set1;
 
  // Create the trace attribute:
  posix_trace_attr_init(&attr);
  posix_trace_attr_setstreamfullpolicy(&attr, POSIX_TRACE_UNTIL_FULL);
  posix_trace_attr_setmaxdatasize(&attr, MY_EVENT_DATA_SIZE);
  posix_trace_attr_setstreamsize(&attr, MY_STREAM_SIZE);

  // Create the stream (without log)
  if ((error = posix_trace_create(0, &attr, &trid)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_create returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  // Create a new event type:
  if ((error = posix_trace_eventid_open ("user event string", 
					 &event_string_id)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_eventid_open returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  // Filter all *system* event types:
  posix_trace_eventset_fill(&set1, POSIX_TRACE_SYSTEM_EVENTS);

  if ((error = posix_trace_set_filter(trid, &set1, 
				      POSIX_TRACE_SET_EVENTSET)) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_get_filter returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }


  // Start the tracing of events: 
  if ((error = posix_trace_start(trid)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_start returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  
  /*
    The three retrieval functions are tested in this way: 

    1) Task "reader_try" tests the trygetnext_event function.
       It executes this retrieval function twice with only one 
       event in the stream: it should retrieve the event and 
       then fail to retrieve the second time.

    2) Task "reader_timed" tests the timedgetnext_event function.
       It executes this retrieval function twice with only one 
       event in the stream: it should retrieve the event and 
       then fail (timeout) to retrieve the second time.

    3) Task "reader_get" tests the getnext_event function.
       It executes this retrieval function twice with only one 
       event in the stream: it should retrieve the event and 
       then fail (suspend forever) to retrieve the second time.
  */
  

  // Create first task:
  posix_trace_event(event_string_id, MY_STRING, sizeof(MY_STRING));
  pthread_create (&reader1, NULL, reader_try, 0); 
  if (test_failed) {
    return -1;
  }


  // Create second task:
  posix_trace_event(event_string_id, MY_STRING, sizeof(MY_STRING));
  pthread_create (&reader2, NULL, reader_timed, 0); 
  /* Ensure the init_module wastes time while the task produces the
     timeout */
  for (do_nothing = 0; do_nothing < 1000000; do_nothing++)
    /* guess what? do nothing! */ ;

  if (test_failed) {
    return -1;
  } else if (! timeout) {
    rtl_printf("%s: [BROK] :BAD TEST: not enough time to produce timeout\n", 
	       TEST_NAME);
    return -1;
  }

  // Create third task:
  posix_trace_event(event_string_id, MY_STRING, sizeof(MY_STRING));
  pthread_create (&reader3, NULL, reader_get, 0); 

  if (test_failed) {
    if (! woke_first) {
      rtl_printf("%s: [BROK] :getnext_event hasn't wake up with an available event\n", 
		 TEST_NAME);
    }
    return -1;
  }
 
 
  // Stop the tracing of events: 
  if ((error = posix_trace_stop(trid)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_stop returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  init_module_finished = 1;

  if (test_failed == 0) {
    rtl_printf("%s: PASS :Online trace/retrieve API works OK\n", 
	       TEST_NAME);  
  }

  return test_failed;

}





void cleanup_module(void) {

  posix_trace_shutdown(trid);

  pthread_cancel(reader1);
  pthread_cancel(reader2);
  pthread_cancel(reader3);

  pthread_join(reader1,NULL); 
  pthread_join(reader2,NULL); 
  pthread_join(reader3,NULL); 

}
